@@ -320,12 +320,6 @@ char *xgetcwd PARAMS((void));
 #define VOID char
 #endif
 
-VOID *xmalloc PARAMS((size_t n));
-VOID *xrealloc PARAMS((VOID *p, size_t n));
-
-/* xstrdup.c */
-char *xstrdup PARAMS((char *string));
-
 /* find global function declarations.  */
 
 /* find.c */
@@ -345,6 +339,7 @@ void set_follow_state PARAMS((enum SymlinkOption opt));
 /* fstype.c */
 char *filesystem_type PARAMS((const char *path, const char *relpath, const struct stat *statp));
 char * get_mounted_filesystems (void);
+dev_t * get_mounted_devices PARAMS((size_t *));
 
 /* parser.c */
 PFB find_parser PARAMS((char *search_name));
