@@ -34,6 +34,7 @@
 #else
 #include <sys/file.h>
 #endif
+#include "../gnulib/lib/xalloc.h"
 #include "../gnulib/lib/human.h"
 #include "../gnulib/lib/canonicalize.h"
 #include <modetype.h>
@@ -62,7 +63,7 @@
   (*(node)->pred_func)((pathname), (stat_buf_ptr), (node))
 
 
-static void init_mount_point_list(void);
+static void init_mounted_dev_list(void);
 static void process_top_path PARAMS((char *pathname));
 static int process_path PARAMS((char *pathname, char *name, boolean leaf, char *parent));
 static void process_dir PARAMS((char *pathname, char *name, int pathlen, struct stat *statp, char *parent));
@@ -169,6 +170,22 @@ enum TraversalDirection
     TraversingDown
   };
 
+
+static int
+following_links(void)
+{
+  switch (symlink_handling)
+    {
+    case SYMLINK_ALWAYS_DEREF:
+      return 1;
+    case SYMLINK_DEREF_ARGSONLY:
+      return (curdepth == 0);
+    case SYMLINK_NEVER_DEREF:
+    default:
+      return 0;
+    }
+}
+
 
 /* optionh_stat() implements the stat operation when the -H option is
  * in effect.
@@ -339,7 +356,7 @@ main (int argc, char **argv)
 
   set_follow_state(SYMLINK_NEVER_DEREF); /* The default is equivalent to -P. */
 
-  init_mount_point_list();
+  init_mounted_dev_list();
 
 #endif
 
@@ -538,25 +555,10 @@ specific_dirname(const char *dir)
     }
 }
 
-enum MountPointStateChange
-  {
-    MountPointRecentlyMounted,
-    MountPointRecentlyUnmounted,
-    MountPointStateUnchanged
-  };
 
-static char *mount_points = NULL;
 
-/* Initialise our idea of what the list of mount points is. 
- * this function is called exactly once.
- */
-static void
-init_mount_point_list(void)
-{
-  assert(NULL == mount_points);
-  mount_points = get_mounted_filesystems();
-}
 
+#if 0
 /* list_item_present: Search for NEEDLE in HAYSTACK.
  *
  * NEEDLE is a normal C string.  HAYSTACK is a list of concatenated C strings, 
@@ -589,6 +591,29 @@ list_item_present(const char *needle, const char *haystack)
   return 0;
 }
 
+static char *mount_points = NULL;
+
+
+/* Initialise our idea of what the list of mount points is. 
+ * this function is called exactly once.
+ */
+static void
+init_mount_point_list(void)
+{
+  assert(NULL == mount_points);
+  mount_points = get_mounted_filesystems();
+}
+
+static void
+refresh_mount_point_list(void)
+{
+  if (mount_points)
+    {
+      free(mount_points);
+      mount_points = NULL;
+    }
+  init_mount_point_list();
+}
 
 /* Determine if a directory has recently had a filesystem 
  * mounted on it or unmounted from it.
@@ -599,13 +624,7 @@ get_mount_point_state(const char *dir)
   int was_mounted, is_mounted;
 
   was_mounted = list_item_present(dir, mount_points);
-  
-  if (mount_points)
-    {
-      free(mount_points);
-    }
-  mount_points = get_mounted_filesystems();
-  
+  refresh_mount_point_list();
   is_mounted = list_item_present(dir, mount_points);
 
   if (was_mounted == is_mounted)
@@ -615,6 +634,76 @@ get_mount_point_state(const char *dir)
   else 
     return MountPointRecentlyUnmounted;
 }
+#endif
+
+
+
+static dev_t *mounted_devices = NULL;
+static size_t num_mounted_devices = 0u;
+
+
+static void
+init_mounted_dev_list()
+{
+  assert(NULL == mounted_devices);
+  assert(0 == num_mounted_devices);
+  mounted_devices = get_mounted_devices(&num_mounted_devices);
+}
+
+static void
+refresh_mounted_dev_list(void)
+{
+  if (mounted_devices)
+    {
+      free(mounted_devices);
+      mounted_devices = 0;
+    }
+  num_mounted_devices = 0u;
+  init_mounted_dev_list();
+}
+
+
+/* Search for device DEV in the array LIST, which is of size N. */
+static int
+dev_present(dev_t dev, const dev_t *list, size_t n)
+{
+  if (list)
+    {
+      while (n-- > 0u)
+	{
+	  if ( (*list++) == dev )
+	    return 1;
+	}
+    }
+  return 0;
+}
+
+enum MountPointStateChange
+  {
+    MountPointRecentlyMounted,
+    MountPointRecentlyUnmounted,
+    MountPointStateUnchanged
+  };
+
+
+
+static enum MountPointStateChange
+get_mount_state(dev_t newdev)
+{
+  int new_is_present, new_was_present;
+  
+  new_was_present = dev_present(newdev, mounted_devices, num_mounted_devices);
+  refresh_mounted_dev_list();
+  new_is_present  = dev_present(newdev, mounted_devices, num_mounted_devices);
+  
+  if (new_was_present == new_is_present)
+    return MountPointStateUnchanged;
+  else if (new_is_present)
+    return MountPointRecentlyMounted;
+  else
+    return MountPointRecentlyUnmounted;
+}
+
 
 
 /* Examine the results of the stat() of a directory from before we
@@ -653,18 +742,21 @@ wd_sanity_check(const char *thing_to_stat,
 		struct stat *newinfo,
 		int parent,
 		int line_no,
-		enum TraversalDirection direction)
+		enum TraversalDirection direction,
+		boolean *changed) /* output parameter */
 {
   const char *fstype;
   char *specific_what = NULL;
-  
   int isfatal = 1;
   
+  *changed = false;
+  
   if ((*xstat) (".", newinfo) != 0)
     error (1, errno, "%s", thing_to_stat);
   
   if (oldinfo->st_dev != newinfo->st_dev)
     {
+      *changed = true;
       specific_what = specific_dirname(what);
       
       /* This condition is rare, so once we are here it is 
@@ -688,7 +780,7 @@ wd_sanity_check(const char *thing_to_stat,
 	   * is much rarer, as it relies on an automounter timeout
 	   * occurring at exactly the wrong moment.
 	   */
-	  enum MountPointStateChange transition = get_mount_point_state(specific_what);
+	  enum MountPointStateChange transition = get_mount_state(newinfo->st_dev);
 	  switch (transition)
 	    {
 	    case MountPointRecentlyUnmounted:
@@ -739,6 +831,7 @@ wd_sanity_check(const char *thing_to_stat,
   /* Device number was the same, check if the inode has changed. */
   if (oldinfo->st_ino != newinfo->st_ino)
     {
+      *changed = true;
       specific_what = specific_dirname(what);
       fstype = filesystem_type(thing_to_stat, ".", newinfo);
       
@@ -754,6 +847,68 @@ wd_sanity_check(const char *thing_to_stat,
       free(specific_what);
     }
 }
+
+enum SafeChdirStatus
+  {
+    SafeChdirOK,
+    SafeChdirFailSymlink,
+    SafeChdirFailStat,
+    SafeChdirFailDefect
+  };
+
+/* Safely perform a change in directory. */
+static int 
+safely_chdir(const char *dest, enum TraversalDirection direction)
+{
+  struct stat statbuf_dest, statbuf_arrived;
+  int rv=SafeChdirFailDefect, dotfd=-1;
+  char *name = NULL;
+
+  errno = 0;
+  dotfd = open(".", O_RDONLY);
+  if (dotfd >= 0)
+    {
+      /* Stat the directory we're going to. */
+      if (0 == (following_links() ? stat : lstat)(dest, &statbuf_dest))
+	{
+#ifdef S_ISLNK
+	  if (!following_links() && S_ISLNK(statbuf_dest.st_mode))
+	    {
+	      rv = SafeChdirFailSymlink;
+	      goto fail;
+	    }
+#endif	  
+	  chdir(dest);
+	}
+      else
+	{
+	  rv = SafeChdirFailStat;
+	  name = specific_dirname(dest);
+	  goto fail;
+	}
+    }
+  else
+    {
+    }
+  
+ fail:
+  if (errno)
+    {
+      if (NULL == name)
+	name = specific_dirname(".");
+      error(0, errno, "%s", name);
+    }
+  
+  free(name);
+  name = NULL;
+  
+  if (dotfd >= 0)
+    {
+      close(dotfd);
+      dotfd = -1;
+    }
+  return rv;
+}
 
 
 /* Safely go back to the starting directory. */
@@ -761,7 +916,8 @@ static void
 chdir_back (void)
 {
   struct stat stat_buf;
-
+  boolean dummy;
+  
   if (starting_desc < 0)
     {
       if (chdir (starting_dir) != 0)
@@ -770,7 +926,7 @@ chdir_back (void)
       wd_sanity_check(starting_dir,
 		      program_name, starting_dir,
 		      &starting_stat_buf, &stat_buf, 0, __LINE__,
-		      TraversingUp);
+		      TraversingUp, &dummy);
     }
   else
     {
@@ -785,7 +941,8 @@ static void
 process_top_path (char *pathname)
 {
   struct stat stat_buf, cur_stat_buf;
-
+  boolean dummy;
+  
   curdepth = 0;
   path_length = strlen (pathname);
 
@@ -809,7 +966,7 @@ process_top_path (char *pathname)
       wd_sanity_check(pathname, program_name,
 		      ".",
 		      &stat_buf, &cur_stat_buf, 0, __LINE__,
-		      TraversingDown);
+		      TraversingDown, &dummy);
 
       process_path (pathname, ".", false, ".");
       chdir_back ();
@@ -992,14 +1149,28 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
       /* Check that we are where we should be. */
       if (1)
 	{
-	  struct stat tmp = {0};
+	  struct stat tmp;
+	  boolean changed;
+	  
+	  memset(&tmp, 0, sizeof(tmp));
 	  tmp.st_dev = dir_ids[dir_curr].dev;
 	  tmp.st_ino = dir_ids[dir_curr].ino;
+	  changed = false;
 	  wd_sanity_check(pathname,
 			  program_name,
 			  ".",
 			  &tmp, &stat_buf, 0, __LINE__, 
-			  TraversingDown);
+			  TraversingDown, &changed);
+	  if (changed)
+	    {
+	      /* If there had been a change but wd_sanity_check()
+	       * accepted it, we need to accept that on the 
+	       * way back up as well, so modify our record 
+	       * of what we think we should see later. 
+	       */
+	      dir_ids[dir_curr].dev = stat_buf.st_dev;
+	      dir_ids[dir_curr].ino = stat_buf.st_ino;
+	    }
 	}
 
       for (namep = name_space; *namep; namep += file_len - pathname_len + 1)
@@ -1043,13 +1214,7 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	  /* We could go back and do the next command-line arg
 	     instead, maybe using longjmp.  */
 	  char const *dir;
-	  boolean deref;
-	  if (symlink_handling == SYMLINK_ALWAYS_DEREF)
-	    deref = true;
-	  else if (symlink_handling == SYMLINK_DEREF_ARGSONLY && (curdepth == 0))
-	    deref = true;
-	  else
-	    deref = false;
+	  boolean deref = following_links() ? true : false;
 	  
 	  if (!deref)
 	    dir = "..";
@@ -1065,9 +1230,11 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	  /* Check that we are where we should be. */
 	  if (1)
 	    {
+	      boolean changed = false;
 	      struct stat tmp;
 	      int problem_is_with_parent;
 	      
+	      memset(&tmp, 0, sizeof(tmp));
 	      if (dir_curr > 0)
 		{
 		  tmp.st_dev = dir_ids[dir_curr-1].dev;
@@ -1082,10 +1249,10 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	      problem_is_with_parent = deref ? 1 : 0;
 	      wd_sanity_check(pathname,
 			      program_name,
-			      deref ? parent : starting_dir,
+			      parent,
 			      &tmp, &stat_buf,
 			      problem_is_with_parent, __LINE__, 
-			      TraversingUp);
+			      TraversingUp, &changed);
 	    }
 	}
 
