@@ -45,8 +45,7 @@
    return NULL if there are no more elements.  */
 
 char *
-next_element (new_path)
-     char *new_path;
+next_element (char *new_path, int curdir_ok)
 {
   static char *path = NULL;	/* Freshly allocated copy of NEW_PATH.  */
   static char *end;		/* Start of next element to return.  */
@@ -67,7 +66,7 @@ next_element (new_path)
       if (final_colon)
 	{
 	  final_colon = 0;
-	  return ".";
+	  return curdir_ok ? "." : NULL;
 	}
       return NULL;
     }
@@ -80,7 +79,7 @@ next_element (new_path)
     {
       /* An empty path element.  */
       *end++ = '\0';
-      return ".";
+      return curdir_ok ? "." : NULL;
     }
   else if (end == NULL)
     {
